unit DaemonUnit1;

// -------------------------------------------------------------------------------
// Demo application on how to use the TDaemon application type
// 2/2022 by arminlinder@armnninlinder.de

// Based on document
// "Taming the daemon: Writing cross-platform service applications in FPC/Lazarus"
// by Michaël Van Canneyt, February 4, 2007
// -------------------------------------------------------------------------------

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, DaemonApp;

const
  DAEMON_VERSION = '1.0';       // Just for logging

type

  { TDaemonThread }

  // This is the "workhorse" of the daemon
  // Execute holds the main work code of the service
  // Do not try the execute method of TDaemon, since it does not multitask,
  // the service thread will stop responding to control messages if looping in the
  // TDaemon execute method. Thus we need a worker thread.

  TDaemonThread = class(TThread)
  private
  public
    procedure Execute; override;  // the worker thread
    constructor Create;
    destructor Destroy; override;
  end;

  { TDaemon1 }

  // The main control structure of the thread
  // Note: there is an execute procedure defined in TDaemon, but you will
  // normally never use it, but create a worker thread

  TDaemon1 = class(TDaemon)
    // Events called through -install and -uninstall
    procedure DataModuleAfterInstall(Sender: TCustomDaemon);
    procedure DataModuleAfterUnInstall(Sender: TCustomDaemon);
    // Service start and stop
    procedure DataModuleStart(Sender: TCustomDaemon; var OK: boolean);
    procedure DataModuleStop(Sender: TCustomDaemon; var OK: boolean);
    // Service pause and resume
    procedure DataModulePause(Sender: TCustomDaemon; var OK: boolean);
    procedure DataModuleContinue(Sender: TCustomDaemon; var OK: boolean);
    // Construction and destruction
    procedure DataModuleCreate(Sender: TObject);
    procedure DataModuleDestroy(Sender: TObject);
  private
    // The worker task, required to keep the service responding to control signals while executing
    FDaemonThread: TDaemonThread;
  public
  end;

var
  Daemon1: TDaemon1;

implementation

// ---------------------------------------------------------------------
// Quick and dirty write log message to file
// Note: TDaemonApplication has extensive logging capabilities built-in!
// ---------------------------------------------------------------------

procedure LogToFile(aMessage: string);

var
  f: Text;
  LogFilePath: string;

  function TimeStamped(S: string): string;

    // Return a timestamped copy of a string

  begin
    Result := FormatDateTime('hh:mm:ss', now) + ' ' + S;
  end;

begin
  // create a new log file in the .exe directory every hour
  LogFilePath := IncludeTrailingPathDelimiter(ExtractFilePath(Application.ExeName)) +
    FormatDateTime('YYYYMMDDhh', now) + '.log';
  AssignFile(f, LogFilePath);
  try
    if FileExists(LogFilePath) then
      Append(f)
    else
    begin
      Rewrite(f);
      writeln(f, TimeStamped('Log created'));
    end;
    Writeln(f, TimeStamped(aMessage));
  finally
    CloseFile(f);
  end;
end;

{$R *.lfm}

{ TDaemonThread }

// ----------------------------------------------------------------------
// The thread providing the execute method of the worker thread.
// ----------------------------------------------------------------------

procedure TDaemonThread.Execute;

// This is the main workhorse, the routine which does
// actually hold the service working code

var
  i: integer;

begin
  LogToFile('Daemon worker thread executing');
  while not Terminated do
  begin
    // placeholder, put your actual service code here
    // ...
    LogToFile('Daemon worker thread running');
    // Thread- and CPUload friendly 5s delay loop
    for i := 1 to 50 do
    begin
      if Terminated then break;
      Sleep(100);
    end;
    // ...
    // ----------------------------------------
  end;
  LogToFile('Daemon worker thread terminated');
end;

// Construction and destruction of the worker thread

constructor TDaemonThread.Create;

begin
  // Create a suspended worker thread
  // The actual thread execution will happen if the OS sends a "Start" Signal
  // See OnStart/DataModelStart event handler
  inherited Create(True);
  LogToFile('Daemon worker thread created');
end;

destructor TDaemonThread.Destroy;

begin
  // Nothing to do here, just for logging
  LogToFile('Daemon worker thread destroyed');
  inherited Destroy;
end;

{ TDaemon1 }

// ---------------------------------------------------------
// This is the main scaffolding of the thread, containing
// the event handlers responding to control signals from the
// operating system service control facility
// ---------------------------------------------------------

// installation and uninstallation events

procedure TDaemon1.DataModuleAfterInstall(Sender: TCustomDaemon);

begin
  // Nothing to do here, just for logging
  LogToFile(Format('Daemon %s installed', [DAEMON_VERSION]));
end;

procedure TDaemon1.DataModuleAfterUnInstall(Sender: TCustomDaemon);

begin
  // Nothing to do here, just for logging
  LogToFile(Format('Daemon %s uninstalled', [DAEMON_VERSION]));
end;

// Service pause and continue signals

procedure TDaemon1.DataModulePause(Sender: TCustomDaemon; var OK: boolean);

begin
  // Nothing to do here, just for logging
  LogToFile('Daemon paused');
end;

procedure TDaemon1.DataModuleContinue(Sender: TCustomDaemon; var OK: boolean);

begin
  // Nothing to do here, just for logging
  LogToFile('Daemon continuing');
end;

// Service start and stop signals

procedure TDaemon1.DataModuleStart(Sender: TCustomDaemon; var OK: boolean);

begin
  OK := True;
  LogToFile('Daemon received start signal');
  // Create a suspended worker thread
  FDaemonThread := TDaemonThread.Create;
  // Parametrize it
  FDaemonThread.FreeOnTerminate := False;
  // Start the worker
  FDaemonThread.Start;
end;

procedure TDaemon1.DataModuleStop(Sender: TCustomDaemon; var OK: boolean);

var
  i: integer;

begin
  LogToFile('Daemon received stop signal');
  // stop and terminate the worker
  FDaemonThread.Terminate;
  // Wait for the thread to terminate.
  FDaemonThread.WaitFor;
  LogToFile('Daemon stopped');
  OK := True;
end;

// Construction and destruction of the daemon class

procedure TDaemon1.DataModuleCreate(Sender: TObject);

begin
  // Nothing to do here, just for logging
  LogToFile(Format('Daemon object %s created', [DAEMON_VERSION]));
end;

procedure TDaemon1.DataModuleDestroy(Sender: TObject);

begin
  // Nothing to do here, just for logging
  LogToFile('Daemon object destroyed');
end;

// --------------------------------
// Global unit init routine
// --------------------------------

procedure RegisterDaemon;

begin
  RegisterDaemonClass(TDaemon1);
end;

initialization
  // Register the daemon with the TCustomDaemonApplication application
  RegisterDaemon;
end.
